@@ -1,14 +1,30 @@
 import numpy as np  # For arrays  (Works well with multiprocessing and mmap)
 from typing import Any, Iterable, Union, List  # To make stuff look nicer
-from types import MethodType  # Used to attach functions to the class
 
 # Custom Classes:
-from .pymadClasses import madObject, madFunction, madReference
+from .pymadClasses import madReference
 from .mad_process import mad_process
 
 # TODO: Make it so that MAD does the loop for variables not python (speed)
 
 
+class last_counter():
+    def __init__(self, size: int, mad_process):
+        self.counter = list(range(size, 0, -1)) 
+        self.mad_process = mad_process
+        self.max_used = 0
+
+    def get(self):
+        assert len(self.counter) != 0, "Assigned too many anonymous, variables, increase num_temp_vars or assign the variables into MAD"
+        val = self.counter.pop()
+        if val > self.max_used:
+            self.max_used = val
+        return f"__last__[{val}]"
+    
+    def set(self, idx):
+        self.counter.append(idx)
+    
+
 class MAD(object):  # Review private and public
     """An object that allows communication with MAD-NG
 
@@ -16,7 +32,7 @@ class MAD(object):  # Review private and public
         py_name: A string indicating the name of the reference to python from MAD-NG, for communication from MAD-NG to Python.
     """
 
-    def __init__(self, py_name: str = "py", mad_path: str = None, debug: bool = False, ipython_use_jedi: bool = False):
+    def __init__(self, py_name: str = "py", mad_path: str = None, debug: bool = False, num_temp_vars: int = 8, ipython_use_jedi: bool = False):
         """Create a MAD Object to interface with MAD-NG.
 
         The modules MADX, elements, sequence, mtable, twiss, beta0, beam, survey, object, track, match are imported into
@@ -29,14 +45,17 @@ class MAD(object):  # Review private and public
                 (default = None)
             debug (bool): Sets debug mode on or off
                 (default = False)
+            num_temp_vars (int): The number of unique temporary variables you intend to use, see :doc:`Managing References <ex-managing-refs>`
+                (default = 255)
             ipython_use_jedi (bool): Allow ipython to use jedi in tab completion, will be slower and may result in MAD-NG throwing errors
                 (default = False)
 
         Returns:
             A MAD object, allowing for communication with MAD-NG
         """
-        #Stop jedi running getattr on my classes...
         self.ipython_use_jedi = ipython_use_jedi
+        self.__last_counter = last_counter(num_temp_vars, self)
+        #Stop jedi running getattr on my classes...
         if not ipython_use_jedi: 
             try:
                 shell = get_ipython().__class__.__name__
@@ -52,6 +71,7 @@ class MAD(object):  # Review private and public
             else return {a, b, ...}
             end
         end
+        __last__ = {}
         """
         )
         self.py_name = py_name
@@ -102,6 +122,9 @@ class MAD(object):  # Review private and public
 
         Returns:
             Data from MAD-NG with type str/int/float/ndarray/bool/list, depending what was asked from MAD-NG.
+
+        Raises:
+            TypeError: If you forget to give a name when receiving a reference,  
         """
         return self.__process.recv(varname)
 
@@ -172,16 +195,11 @@ class MAD(object):  # Review private and public
     def __errhdlr(self, on_off: bool):
         self.send(f"py:__err("+ str(on_off).lower() + ")")
 
-    def __safe_send_recv(func):
-        def safe_send_recv(self, *args):
-            self.__errhdlr(True)
-            res = func(self, *args)
-            self.__errhdlr(False)
-            return res
-        return safe_send_recv
+    def __safe_send(self, string: str):
+        return self.send(f"py:__err(true); {string}; py:__err(false);")
 
     def load(self, module: str, vars: List[str] = []):
-        """Import modules into the MAD-NG environment
+        """Load modules into the MAD-NG environment
 
         Retrieve the classes in MAD-NG from the module ``module``, while only importing the classes in the list ``vars``.
         If no list is provided, it is assumed that you would like to import every class from the module.
@@ -201,7 +219,7 @@ class MAD(object):  # Review private and public
         self.__process.send(script)
 
     def __getattr__(self, item):
-        if item[0] == "_" and not item == "__last__":
+        if item[0] == "_" and not item[:8] == "__last__":
             raise AttributeError(item)
         return self.recv_vars(item)
 
@@ -241,8 +259,8 @@ class MAD(object):  # Review private and public
         Returns:
             The evaluated result.
         """
-        self.__process.send("__last__ =" + input)
-        return self["__last__"]
+        last_name = self.__last_counter.get()
+        return self.send(f"{last_name} =" + input)[last_name]
 
     def MADX_env_send(self, input: str):
         """Open the MAD-X environment in MAD-NG and directly send code.
@@ -252,6 +270,20 @@ class MAD(object):  # Review private and public
         """
         return self.__process.send("MADX:open_env()\n" + input + "\nMADX:close_env()")
 
+    def py_strs_to_mad_strs(self, input: Union[str, List[str]]):
+        """Add ' to either side of a string or list of strings
+        
+        Args: 
+            input(str/list[str]): The string or list of strings that you would like to add ' either side to.
+
+        Returns:
+            A string or list of strings with ' placed at the beginning and the end of the string.
+        """
+        if isinstance(input, list):
+            return ["'" + x + "'" for x in input]
+        else:
+            return "'" + input + "'"
+
     # ----------------------------------------------------------------------------------------------#
 
     # ----------------------------------Sending variables across to MAD----------------------------------------#
@@ -279,12 +311,12 @@ class MAD(object):  # Review private and public
         else:
             assert isinstance(vars, list), "A list of names must be matched with a list of variables"
             assert len(vars) == len(names), "The number of names must match the number of variables"
-        for i in range(len(vars)):
+        for i, var in enumerate(vars):
             if isinstance(vars[i], madReference):
-                self.__process.send(f"{names[i]} = {vars[i].__name__}")
+                self.__process.send(f"{names[i]} = {var.__name__}")
             else:
                 self.__process.send(f"{names[i]} = {self.py_name}:recv()")
-                self.__process.send(vars[i])
+                self.__process.send(var)
     # -------------------------------------------------------------------------------------------------------------#
 
     # -----------------------------------Receiving variables from to MAD-------------------------------------------#
@@ -306,10 +338,10 @@ class MAD(object):  # Review private and public
             lst_rtn = True
 
         returnVars = []
-        for i in range(len(names)):
-            if names[i][:2] != "__" or "__last__" in names[i]:  # Check for private variables
-                self.__process.send(f"{self.py_name}:__err(true):send({names[i]}):__err(false)")
-                returnVars.append(self.__process.recv(names[i]))
+        for name in names:
+            if name[:2] != "__" or name[:8] == "__last__":  # Check for private variables
+                self.__process.send(f"{self.py_name}:__err(true):send({name}):__err(false)")
+                returnVars.append(self.__process.recv(name))
         
         if lst_rtn:
             return tuple(returnVars)
@@ -320,80 +352,64 @@ class MAD(object):  # Review private and public
 
     # ----------------------------------Calling/Creating functions-------------------------------------------------#
     def __call_func(self, funcName: str, *args):
-        """Call the function funcName and store the result in ``__last__``.
-
-        To assign the return values names, then use ``mad[names] = mad.call_func(funcName, args)``, where ``names`` is a
-        variable length argument list of strings.
-
-        Args:
-            funcName (str): The name of the function that you would like to call in MAD-NG.
-            *args: Variable length argument list required to call the function funcName.
-
-        Returns:
-            A reference to a list of the return values of the function.
-        """
+        """Call the function funcName and store the result in ``__last__``."""
+        last_name = self.__last_counter.get()
+        args_string, vars_to_send = self.__get_args_string(*args)
         self.__process.send(
-            f"__last__ = __mklast__({self.__to_MAD_string(funcName)}({self.__getArgsAsString(*args)}))\n"
+            f"{last_name} = __mklast__({funcName}({args_string}))\n"
         )
-        return madReference("__last__", self)
+
+        for var in vars_to_send:
+            self.send(var)
+        return madReference(last_name, self)
     # -------------------------------------------------------------------------------------------------------------#
 
     # -------------------------------String Conversions--------------------------------------------------#
-    def __getKwargAsString(self, **kwargs):
+    def __get_kwargs_string(self, **kwargs):
         # Keep an eye out for failures when kwargs is empty, shouldn't occur in current setup
-        """Convert a keyword argument input to a string used by MAD-NG
-        THIS NEEDS TO BE IMPROVED!, WHAT IF USER WANTS BIG DATA AS ARGUMENT"""
+        """Convert a keyword argument input to a string used by MAD-NG"""
         kwargsString = "{"
+        vars_to_send = []
         for key, item in kwargs.items():
             keyString = str(key).replace("'", "")
-            itemString = self.__to_MAD_string(item)
+            itemString, var = self.__to_MAD_string(item)
             kwargsString += keyString + " = " + itemString + ", "
-        return kwargsString + "}"
+            vars_to_send.extend(var)
+        return kwargsString + "}", vars_to_send
 
-    def __getArgsAsString(self, *args):
-        """Convert an argument input to a string used by MAD-NG
-        THIS NEEDS TO BE IMPROVED!, WHAT IF USER WANTS BIG DATA AS ARGUMENT"""
-        argStr = ""
+    def __get_args_string(self, *args):
+        """Convert an argument input to a string used by MAD-NG"""
+        mad_string = ""
+        vars_to_send = []
         for arg in args:
-            argStr += self.__to_MAD_string(arg) + ", "
-        return argStr[:-2]  # Assumes args is always put last
+            string, var = self.__to_MAD_string(arg)
+            mad_string += string + ", "
+            vars_to_send.extend(var)
+        return mad_string[:-2], vars_to_send
 
-    def __pyToLuaLst(self, lst: List[Any]) -> str:
-        """Convert a python list to a lua list in a string, used when sending information to MAD-NG, should not need to be accessed by user"""
-        luaString = "{"
-        for item in lst:
-            luaString += self.__to_MAD_string(item, True) + ", "
-        return luaString + "}"
-
-    def __to_MAD_string(self, var: Any, convertString=False):
-        """Convert a list of objects into the required string for MAD-NG.
-
-        Args:
-            var (Any): A variable that you would like converted into the string that can be used by MAD-NG
-            convertString (bool): A boolean that when true converts ``var = "Hello World"`` to ``"'Hello World'"`` (Converts to string, not expression in MAD-NG)
-                (default = False)
-
-        Returns:
-            The converted string, ready for use, directly in the MAD-NG environment
-
-        """
-        if isinstance(var, (list, np.ndarray)):
-            return self.__pyToLuaLst(var)
-        elif var == None:
-            return "nil"
-        elif isinstance(var, str) and convertString:
-            return "'" + var + "'"
+    def __to_MAD_string(self, var: Any):
+        """Convert a list of objects into the required string for MAD-NG. 
+        Converting string instead of sending more data is up to 2x faster (therefore last resort)."""
+        if isinstance(var, list):
+            string, vars_to_send = self.__get_args_string(*var)
+            return "{" + string + "}", vars_to_send
+        elif var is None:
+            return "nil", []
+        elif isinstance(var, str):
+            return var, []
         elif isinstance(var, complex):
-            # Remove brackets and convert j to i
-            return str(var)[1:-1].replace("j", "i") 
+            string = str(var)
+            return (string[0] == "(" and string[1:-1] or string).replace("j", "i"), []
         elif isinstance(var, (madReference)):
-            return var.__name__
+            return var.__name__, []
         elif isinstance(var, dict):
-            return self.__getKwargAsString(**var)
-        elif callable(var):
-            return var.__name__
+            return self.__get_kwargs_string(**var)
+        elif isinstance(var, bool):
+            return str(var).lower(), []
+        elif isinstance(var, (float, int)):
+            return str(var), []
         else:
-            return str(var).replace("False", "false").replace("True", "true")
+            return f"{self.py_name}:recv()", [var]
 
     # ---------------------------------------------------------------------------------------------------#
 
@@ -410,10 +426,14 @@ class MAD(object):  # Review private and public
         Returns:
             A reference to the deffered expression object.
         """
+        last_name = self.__last_counter.get()
+        kwargs_string, vars_to_send = self.__get_kwargs_string(**kwargs)
         self.__process.send(
-            f"__last__ = __mklast__( MAD.typeid.deferred {{ {self.__getKwargAsString(**kwargs).replace('=', ':=')[1:-3]} }} )"
+            f"{last_name} = __mklast__( MAD.typeid.deferred {{ {kwargs_string.replace('=', ':=')[1:-3]} }} )"
         )
-        return madReference("__last__", self)
+        for var in vars_to_send:
+            self.send(var)
+        return madReference(last_name, self)
 
     def __dir__(self) -> Iterable[str]:
         pyObjs = [x for x in super(MAD, self).__dir__() if x[0] != "_"]
@@ -437,3 +457,5 @@ class MAD(object):  # Review private and public
         del self
 
     # ---------------------------------------------------------------------------------------------------#
+    def __del__(self):
+        print("max references used:", self.__last_counter.max_used)
\ No newline at end of file
